@@ -17,6 +17,7 @@ from trytond.transaction import Transaction, check_access, without_check_access
 
 from . import fields
 from .fields import on_change_result
+from .fields.field import _set_value
 from .model import Model
 
 __all__ = ['ModelView']
@@ -822,7 +823,13 @@ class ModelView(Model):
             func.change = set(fields)
             if methods:
                 func.change_methods = set(methods)
-            return func
+
+            @wraps(func)
+            def wrapper(self, *args, **kwargs):
+                for field in fields:
+                    _set_value(self, field)
+                return func(self, *args, **kwargs)
+            return wrapper
         return decorator
 
     @on_change
