@@ -1,23 +1,27 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
-import dateutil.tz
-import os
 import inspect
 import logging
 import math
+import mimetypes
+import operator
+import os
+import pathlib
+import shutil
 import subprocess
 import tempfile
 import time
 import unicodedata
 import warnings
 import zipfile
-import operator
-from email.mime.multipart import MIMEMultipart
-from email.mime.text import MIMEText
+from email.message import EmailMessage
+from functools import partial
 from io import BytesIO
 from itertools import groupby
 
+import dateutil.tz
+
 try:
     import html2text
 except ImportError:
@@ -29,20 +33,32 @@ except ImportError:
     weasyprint = None
 
 from genshi.filters import Translator
+from genshi.template.text import TextTemplate
 
-from trytond.i18n import gettext
+from trytond.config import config
+from trytond.i18n import gettext, ngettext
+from trytond.model.exceptions import AccessError
 from trytond.pool import Pool, PoolBase
-from trytond.transaction import Transaction
-from trytond.tools import slugify
-from trytond.url import URLMixin
 from trytond.rpc import RPC
-from trytond.exceptions import UserError
+from trytond.tools import slugify
+from trytond.transaction import Transaction, check_access
+from trytond.url import URLMixin
 
-warnings.simplefilter("ignore")
-import relatorio.reporting  # noqa: E402
-warnings.resetwarnings()
 try:
-    from relatorio.templates.opendocument import Manifest, MANIFEST
+    from trytond.tools import barcode
+except ImportError:
+    barcode = None
+try:
+    from trytond.tools import qrcode
+except ImportError:
+    qrcode = None
+
+with warnings.catch_warnings():
+    warnings.simplefilter("ignore")
+    import relatorio.reporting  # noqa: E402
+
+try:
+    from relatorio.templates.opendocument import MANIFEST, Manifest
 except ImportError:
     Manifest, MANIFEST = None, None
 
@@ -89,7 +105,22 @@ TIMEDELTA_DEFAULT_CONVERTER['w'] = TIMEDELTA_DEFAULT_CONVERTER['d'] * 7
 TIMEDELTA_DEFAULT_CONVERTER['M'] = TIMEDELTA_DEFAULT_CONVERTER['d'] * 30
 TIMEDELTA_DEFAULT_CONVERTER['Y'] = TIMEDELTA_DEFAULT_CONVERTER['d'] * 365
 
-REPORT_NAME_MAX_LENGTH = 200
+NO_BREAKING_SPACE = '\u00A0'
+
+# For most OS maximum filename is 255 but Excel has a limitation which include
+# the path of 218.
+# As on Windows report is most likely to be open from
+# C:\Users\<username>\AppData\Local\Temp\tryton_<random>\ which has a length of
+# 56 with 12 for username and 8 for random. So 162 should be the maximum but we
+# round it to 100.
+REPORT_NAME_MAX_LENGTH = 100
+
+CONVERT_COMMAND = config.get(
+    'report', 'convert_command',
+    default='soffice --headless --nolockcheck --nodefault --norestore '
+    '--convert-to "%(output_extension)s" '
+    '--outdir "%(directory)s" '
+    '"%(input_path)s"')
 
 
 class TranslateFactory:
@@ -98,35 +129,55 @@ class TranslateFactory:
         self.report_name = report_name
         self.translation = translation
 
-    def __call__(self, text):
+    def gettext(self, text):
         return self.translation.get_report(self.report_name, text)
 
+    def ngettext(self, text, text_plural, n):
+        return self.translation.get_report(
+            self.report_name, text, text_plural, n)
+
 
 class Report(URLMixin, PoolBase):
 
     @classmethod
     def __setup__(cls):
-        super(Report, cls).__setup__()
+        super().__setup__()
         cls.__rpc__ = {
             'execute': RPC(),
             }
 
     @classmethod
-    def check_access(cls):
+    def check_access(cls, action, model, ids):
         pool = Pool()
         ActionReport = pool.get('ir.action.report')
         User = pool.get('res.user')
+        Group = pool.get('res.group')
+        ModelAccess = pool.get('ir.model.access')
 
         if Transaction().user == 0:
             return
 
-        groups = set(User.get_groups())
-        report_groups = ActionReport.get_groups(cls.__name__)
-        if report_groups and not groups & report_groups:
-            raise UserError('Calling report %s is not allowed!' % cls.__name__)
+        with check_access():
+            groups = set(User.get_groups())
+            report_groups = ActionReport.get_groups(cls.__name__, action.id)
+            if report_groups and not groups & report_groups:
+                groups = Group.browse(User.get_groups())
+                raise AccessError(
+                    gettext(
+                        'ir.msg_access_report_error',
+                        report=cls.__name__),
+                    gettext(
+                        'ir.msg_context_groups',
+                        groups=', '.join(g.rec_name for g in groups)))
+
+            if model:
+                Model = pool.get(model)
+                ModelAccess.check(model, 'read')
+                # Check read access
+                Model.read(ids, ['id'])
 
     @classmethod
-    def header_key(cls, record):
+    def header_key(cls, record, data):
         return ()
 
     @classmethod
@@ -139,11 +190,12 @@ class Report(URLMixin, PoolBase):
             report type,
             data,
             a boolean to direct print,
-            the report name
+            the report name (with or without the record names)
         '''
         pool = Pool()
         ActionReport = pool.get('ir.action.report')
-        cls.check_access()
+        context = Transaction().context
+        ids = list(map(int, ids))
 
         action_id = data.get('action_id')
         if action_id is None:
@@ -155,6 +207,9 @@ class Report(URLMixin, PoolBase):
         else:
             action_report = ActionReport(action_id)
 
+        model = action_report.model or data.get('model')
+        cls.check_access(action_report, model, ids)
+
         def report_name(records, reserved_length=0):
             names = []
             name_length = 0
@@ -162,8 +217,15 @@ class Report(URLMixin, PoolBase):
             max_length = (REPORT_NAME_MAX_LENGTH
                 - reserved_length
                 - len(str(record_count)) - 2)
+            if action_report.record_name:
+                template = TextTemplate(action_report.record_name)
+            else:
+                template = None
             for record in records[:5]:
-                record_name = record.rec_name
+                if template:
+                    record_name = template.generate(record=record).render()
+                else:
+                    record_name = record.rec_name
                 name_length += len(
                     unicodedata.normalize('NFKD', record_name)) + 1
                 if name_length > max_length:
@@ -175,21 +237,22 @@ class Report(URLMixin, PoolBase):
                 name += '__' + str(record_count - len(names))
             return name
 
-        records = []
-        model = action_report.model or data.get('model')
         if model:
             records = cls._get_records(ids, model, data)
+        else:
+            records = []
 
+        header_key = partial(cls.header_key, data=data)
         if not records:
             groups = [[]]
             headers = [{}]
         elif action_report.single:
             groups = [[r] for r in records]
-            headers = [dict(cls.header_key(r)) for r in records]
+            headers = [dict(header_key(r)) for r in records]
         else:
             groups = []
             headers = []
-            for key, group in groupby(records, key=cls.header_key):
+            for key, group in groupby(records, key=header_key):
                 groups.append(list(group))
                 headers.append(dict(key))
 
@@ -217,12 +280,15 @@ class Report(URLMixin, PoolBase):
         if not isinstance(content, str):
             content = bytearray(content) if bytes == str else bytes(content)
         action_report_name = action_report.name[:REPORT_NAME_MAX_LENGTH]
-        filename = join_string.join(
-            filter(None, [
-                action_report_name,
-                report_name(
-                    records, len(action_report_name) + len(join_string))]))
-        return (oext, content, action_report.direct_print, filename)
+        if context.get('with_rec_name', True):
+            filename = join_string.join(
+                filter(None, [
+                    action_report_name,
+                    report_name(
+                        records, len(action_report_name) + len(join_string))]))
+        else:
+            filename = action_report_name
+        return (oext, content, bool(action_report.direct_print), filename)
 
     @classmethod
     def _execute(cls, records, header, data, action):
@@ -242,6 +308,7 @@ class Report(URLMixin, PoolBase):
 
         class TranslateModel(object):
             _languages = {}
+            __class__ = Model
 
             def __init__(self, id):
                 self.id = id
@@ -292,13 +359,20 @@ class Report(URLMixin, PoolBase):
         report_context['format_timedelta'] = cls.format_timedelta
         report_context['format_currency'] = cls.format_currency
         report_context['format_number'] = cls.format_number
+        report_context['format_number_symbol'] = cls.format_number_symbol
         report_context['datetime'] = datetime
+        if barcode:
+            report_context['barcode'] = cls.barcode
+        if qrcode:
+            report_context['qrcode'] = cls.qrcode
 
         def set_lang(language=None):
             if isinstance(language, Lang):
                 language = language.code
             Transaction().set_context(language=language)
         report_context['set_lang'] = set_lang
+        report_context['msg_gettext'] = gettext
+        report_context['msg_ngettext'] = ngettext
 
         return report_context
 
@@ -308,12 +382,13 @@ class Report(URLMixin, PoolBase):
             pool = Pool()
             Translation = pool.get('ir.translation')
             translate = TranslateFactory(cls.__name__, Translation)
-            translator = Translator(lambda text: translate(text))
+            translator = Translator(translate)
             # Do not use Translator.setup to add filter at the end
             # after set_lang evaluation
             template.filters.append(translator)
             if hasattr(template, 'add_directives'):
                 template.add_directives(Translator.NAMESPACE, translator)
+            return translate
 
     @classmethod
     def render(cls, report, report_context):
@@ -324,8 +399,12 @@ class Report(URLMixin, PoolBase):
             loader = relatorio.reporting.MIMETemplateLoader()
             klass = loader.factories[loader.get_type(mimetype)]
             template = klass(BytesIO(report.report_content))
-            cls._callback_loader(report, template)
             report.set_template_cached(template)
+        translate = cls._callback_loader(report, template)
+        if translate:
+            report_context = report_context.copy()
+            report_context['gettext'] = translate.gettext
+            report_context['ngettext'] = translate.ngettext
         data = template.generate(**report_context).render()
         if hasattr(data, 'getvalue'):
             data = data.getvalue()
@@ -342,45 +421,51 @@ class Report(URLMixin, PoolBase):
                 and output_format == 'pdf'):
             return output_format, weasyprint.HTML(string=data).write_pdf()
 
-        if output_format in MIMETYPES:
+        if input_format == output_format and output_format in MIMETYPES:
             return output_format, data
 
-        dtemp = tempfile.mkdtemp(prefix='trytond_')
-        path = os.path.join(
-            dtemp, report.report_name + os.extsep + input_format)
-        oext = FORMAT2EXT.get(output_format, output_format)
+        directory = tempfile.mkdtemp(prefix='trytond_')
+        input_extension = FORMAT2EXT.get(input_format, input_format)
+        output_extension = FORMAT2EXT.get(output_format, output_format)
+        path = pathlib.Path(directory, report.report_name)
+        input_path = path.with_suffix(os.extsep + input_extension)
+        output_path = path.with_suffix(os.extsep + output_extension)
         mode = 'w+' if isinstance(data, str) else 'wb+'
-        with open(path, mode) as fp:
+        with open(input_path, mode) as fp:
             fp.write(data)
         try:
-            cmd = ['soffice',
-                '--headless', '--nolockcheck', '--nodefault', '--norestore',
-                '--convert-to', oext, '--outdir', dtemp, path]
-            output = os.path.splitext(path)[0] + os.extsep + oext
+            cmd = CONVERT_COMMAND % {
+                'directory': directory,
+                'input_format': input_format,
+                'input_extension': input_extension,
+                'input_path': input_path,
+                'output_format': output_format,
+                'output_extension': output_extension,
+                'output_path': output_path,
+                }
             for count in range(retry, -1, -1):
                 if count != retry:
                     time.sleep(0.02 * (retry - count))
                 try:
-                    subprocess.check_call(cmd, timeout=timeout)
+                    subprocess.check_call(cmd, timeout=timeout, shell=True)
                 except subprocess.CalledProcessError:
                     if count:
                         continue
                     logger.error(
-                        "fail to convert %s to %s", report.report_name, oext,
-                        exc_info=True)
+                        "fail to convert %s to %s",
+                        report.report_name, output_format, exc_info=True)
                     break
-                if os.path.exists(output):
-                    with open(output, 'rb') as fp:
-                        return oext, fp.read()
+                if os.path.exists(output_path):
+                    with open(output_path, 'rb') as fp:
+                        return output_extension, fp.read()
             else:
                 logger.error(
-                    'fail to convert %s to %s', report.report_name, oext)
+                    'fail to convert %s to %s',
+                    report.report_name, output_format)
             return input_format, data
         finally:
             try:
-                os.remove(path)
-                os.remove(output)
-                os.rmdir(dtemp)
+                shutil.rmtree(directory, ignore_errors=True)
             except OSError:
                 pass
 
@@ -447,7 +532,7 @@ class Report(URLMixin, PoolBase):
                 # Add space if no time
                 text += ' '
             text += ('%.6f' % value)[1:]
-        return text
+        return text.replace(' ', NO_BREAKING_SPACE)
 
     @classmethod
     def format_currency(
@@ -457,17 +542,40 @@ class Report(URLMixin, PoolBase):
         Lang = pool.get('ir.lang')
         if lang is None:
             lang = Lang.get()
-        return lang.currency(value, currency, symbol, grouping, digits=digits)
+        return lang.currency(
+            value, currency, symbol=symbol, grouping=grouping, digits=digits)
 
     @classmethod
-    def format_number(cls, value, lang, digits=2, grouping=True,
-            monetary=None):
+    def format_number(
+            cls, value, lang, digits=None, grouping=True, monetary=None):
         pool = Pool()
         Lang = pool.get('ir.lang')
         if lang is None:
             lang = Lang.get()
-        return lang.format('%.' + str(digits) + 'f', value,
-            grouping=grouping, monetary=monetary)
+        return lang.format_number(
+            value, digits=digits, grouping=grouping, monetary=monetary)
+
+    @classmethod
+    def format_number_symbol(
+            cls, value, lang, symbol, digits=None, grouping=True):
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        if lang is None:
+            lang = Lang.get()
+        return lang.format_number_symbol(
+            value, symbol, digits=digits, grouping=grouping)
+
+    if barcode:
+        @classmethod
+        def barcode(cls, name, code, size=(), **kwargs):
+            image = barcode.generate_svg(name, code, **kwargs)
+            return (image, 'image/svg+xml', *size)
+
+    if qrcode:
+        @classmethod
+        def qrcode(cls, code, size=(), **kwargs):
+            image = qrcode.generate_svg(code, **kwargs)
+            return (image, 'image/svg+xml', *size)
 
 
 def get_email(report, record, languages):
@@ -486,23 +594,46 @@ def get_email(report, record, languages):
         Report_ = pool.get(report_name, type='report')
     converter = None
     title = None
-    msg = MIMEMultipart('alternative')
-    msg.add_header('Content-Language', ', '.join(l.code for l in languages))
-    for language in languages:
-        with Transaction().set_context(language=language.code):
+    msg = EmailMessage()
+    header_factory = msg.policy.header_factory
+    for alternative, language in enumerate(languages):
+        with Transaction().set_context(
+                language=language.code, with_rec_name=False):
             ext, content, _, title = Report_.execute(
                 [record.id], {
                     'action_id': report_id,
                     'language': language,
                     })
-        if ext == 'html' and html2text:
+        for map in [mimetypes.types_map, mimetypes.common_types]:
+            if '.' + ext in map:
+                mimetype = map['.' + ext]
+                maintype, subtype = mimetype.split('/')
+                break
+        else:
+            maintype, subtype = 'application', ext
+        if maintype == 'text' and subtype == 'html' and html2text:
             if not converter:
                 converter = html2text.HTML2Text()
-            part = MIMEText(
-                converter.handle(content), 'plain', _charset='utf-8')
-            part.add_header('Content-Language', language.code)
-            msg.attach(part)
-        part = MIMEText(content, ext, _charset='utf-8')
-        part.add_header('Content-Language', language.code)
-        msg.attach(part)
+            content_text = converter.handle(content)
+            msg.add_alternative(content_text, subtype='plain', headers=[
+                    header_factory('Content-Language', language.code),
+                    ])
+        types = {
+            'subtype': subtype,
+            }
+        if not isinstance(content, str):
+            types['maintype'] = maintype
+        if alternative or msg.is_multipart():
+            msg.add_alternative(
+                content, **types, headers=[
+                    header_factory('Content-Language', language.code),
+                    ])
+        else:
+            msg.set_content(
+                content, **types, headers=[
+                    header_factory('Content-Language', language.code),
+                    ])
+    if msg.is_multipart():
+        msg.add_header(
+            'Content-Language', ', '.join(l.code for l in languages))
     return msg, title
